#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

import os
import sys
import functools
import warnings
import datetime
import dateutil.relativedelta
import hues

console = hues.SimpleConsole(stdout=sys.stderr)


def save_cookies(session):
    """Save cookies from a session.

    Arguments:
        session (`requests.Session`): a session with cookies
            to save.

    Note:
        Cookies are saved in a plain text file in the system
        system temporary directory. Use ``tempfile.gettempdir()``
        to find where it is.
    """
    try:
        if not os.path.isdir(os.path.dirname(session.cookies.filename)):
            os.mkdir(os.path.dirname(session.cookies.filename))
        session.cookies.save()
    except IOError:
        pass


def load_cookies(session):
    """Load saved cookies to a session.

    Arguments:
        session (`requests.session`): a session with saved
            cookies to load.

    Note:
        Expired cookies are cleaned, so the session will
        only load active cookies.
    """
    try:
        session.cookies.load()
    except IOError:
        pass
    session.cookies.clear_expired_cookies()


def get_times(timeframe):
    """Get absolute times from a maybe incomplete timeframe.

    Argument:
        timeframe (`tuple`): a couple of `datetime.date` objects
            or eventually ``None``.

    Returns:
        tuple: a tuple with absolute `datetime.date` values (start ``None``
            is replaced with ``datetime.date.today()``, and end ``None``
            is replaced with ``datetime.date.fromtimestamp(0)``, i.e. the
            epoch).

    Raises:
        TypeError: If the provided object is not a sequence of (at least)
            two arguments.
    """
    if timeframe is None:
        timeframe = (None, None)
    try:
        start_time = timeframe[0] or datetime.date.today()
        end_time = timeframe[1] or datetime.date.fromtimestamp(0)
    except (IndexError, AttributeError):
        raise TypeError("'timeframe' must be a tuple of dates !")
    return start_time, end_time


def get_times_from_cli(cli_token):
    """Convert a CLI token to a datetime tuple.

    Argument:
        cli_token (`str`): an isoformat datetime token ([ISO date]:[ISO date])
            or a special value among:
                * thisday
                * thisweek
                * thismonth
                * thisyear

    Returns:
        tuple: a datetime.date objects couple, where the first item is
            the start of a time frame and the second item the end of the
            time frame. Both elements can also be None, if no date was
            provided.

    Raises:
        ValueError: when the CLI token is not in the right format
            (no colon in the token, not one of the special values, dates
            are not in proper ISO-8601 format.)

    See also:
        `ISO-8601 specification <https://en.wikipedia.org/wiki/ISO_8601>`_.
    """

    today = datetime.date.today()

    if cli_token=="thisday":
        return today, today
    elif cli_token=="thisweek":
        return today, today - dateutil.relativedelta.relativedelta(days=7)
    elif cli_token=="thismonth":
        return today, today - dateutil.relativedelta.relativedelta(months=1)
    elif cli_token=="thisyear":
        return today, today - dateutil.relativedelta.relativedelta(years=1)
    else:
        try:
            start_date, stop_date = cli_token.split(':')
        except ValueError:
            raise ValueError("--time parameter must contain a colon (:)")
        if not start_date and not stop_date: # ':', no start date, no stop date
            return None, None
        try:
            start_date = date_from_isoformat(start_date) if start_date else None #datetime.date.today()
            stop_date = date_from_isoformat(stop_date) if stop_date else None #datetime.date.fromtimestamp(0)
        except ValueError:
            raise ValueError("--time parameter was not provided ISO formatted dates")
        return start_date, stop_date


def date_from_isoformat(isoformat_date):
    """Convert an ISO-8601 date into a `datetime.date` object.

    Argument:
        isoformat_date (`string`): a date in ISO-8601 format
            (YYYY-MM-DD)

    Returns:
        `datetime.date`: the date object corresponding to the given ISO
            formatted date.

    Raises:
        ValueError: when the date could not be converted successfully.

    See also:
        `ISO-8601 specification <https://en.wikipedia.org/wiki/ISO_8601>`_.
    """
    year, month, day = isoformat_date.split('-')
    return datetime.date(int(year), int(month), int(day))


def warn_with_hues(message, category, filename, lineno, file=None, line=None):
    """Use `hues` to log warnings.
    """
    console.warn(message)

def warn_windows(message, category, filename, lineno, file=None, line=None):
    """Use a `hues`-like format to log warnings without color.
    """
    print(("{d.hour}:{d.minute}:{d.second} - WARNING - "
          "{msg}").format(d=datetime.datetime.now(), msg=message), file=sys.stderr)

def wrap_warnings(func):
    """Have the function patch `warnings.showwarning` when it is called.
    """
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        showwarning = warnings.showwarning
        if os.name == 'posix':
            warnings.showwarning = warn_with_hues
        else:
            warnings.showwarning = warn_windows
        result = func(*args, **kwargs)
        warnings.showwarning = showwarning
        return result
    return new_func



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

__author__ = "althonos"
__author_email__ = "martin.larralde@ens-cachan.fr"
__version__ = "0.13.1"
__license___ = "GPL-3+"

try:
    from .cli import main
    from .core import InstaLooter
    from . import urlgen
except ImportError:
    pass



# coding: utf-8
"""
instaLooter.urlgen
==================

Contains several url generators that can be passed as ``url_generator``
argument to the ``InstaLooter`` constructor.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import re

__all__ = ["default", "resizer", "thumbnail"]


DEFAULT_RX = re.compile(r'(https://.*?/.*?/).*(/.*\.jpg)')
NEW_DEFAULT_RX = re.compile(r'(https://.*?/h-ak-igx/).*(/.*\.jpg)')
RESIZER_RX = re.compile(r"(s[0-9x]*/)?(e[0-9]*)/")


def default(media):
    """Generates a link to the default picture, without processing effects.
    """
    result = NEW_DEFAULT_RX.search(media['display_src'])

    return 'anything'.join(result.groups()) \
        if result is not None \
      else ''.join(DEFAULT_RX.search(media['display_src']).groups())


def resizer(size):
    """Creates a generator that generate a link to a scaled picture.

    Aspect ratio is not modified, and the largets of the two
    dimensions of the picture will equal to ``size`` once it
    is transformed.
    """
    target = 's{0}x{0}/'.format(size)
    def resize(media):
        return RESIZER_RX.sub(target, media['display_src'])
    resize.__doc__ = "Generates a link to a picture {} pixels large.".format(size)
    return resize


def thumbnail(media):
    """Generates a link to the thumbnail.
    """
    try:
        return media['thumbnail_src']
    except KeyError:
        url_parts = media['display_src'].split('/')
        url_parts.insert(4, 'c121.0.669.669')
        url_parts.insert(5, 's640x640')
        return '/'.join(url_parts)



#!/usr/bin/env python
# coding: utf-8
from __future__ import (
    absolute_import,
    unicode_literals,
)

import atexit
import copy
import json
import os
import datetime
import contextlib
import progressbar
import random
import re
import warnings
import threading
import requests
import six
import time
import tempfile
import bs4 as bs

from .urlgen import default
from .worker import InstaDownloader
from .utils import get_times, save_cookies, load_cookies

PARSER = 'html.parser'


class InstaLooter(object):
    """A brutal Instagram looter that raids without API tokens.
    """

    _RX_SHARED_DATA = re.compile(r'window._sharedData = ({[^\n]*});')
    _RX_TEMPLATE = re.compile(r'{([a-zA-Z]*)}')
    _RX_CODE_URL = re.compile(r'p/([^/]*)')

    URL_HOME = "https://www.instagram.com/"
    URL_LOGIN = "https://www.instagram.com/accounts/login/ajax/"
    URL_LOGOUT = "https://www.instagram.com/accounts/logout/"

    COOKIE_FILE = os.path.join(tempfile.gettempdir(), "instaLooter", "cookies.txt")

    _TEMPLATE_MAP = {
        'id': lambda m: m.get('id'),
        'code': lambda m: m.get('code') or m.get('shortcode'),
        'ownerid': lambda m: m.get('owner', dict()).get('id'),
        'username': lambda m: m.get('owner', dict()).get('username'),
        'fullname': lambda m: m.get('owner', dict()).get('full_name'),
        'datetime': lambda m: ("{0.year}-{0.month}-{0.day} {0.hour}h{0.minute}m{0.second}"
            "s{0.microsecond}".format(datetime.datetime.fromtimestamp(m['date'])))
            if 'date' in m else None,
        'date': lambda m: datetime.date.fromtimestamp(m['date'])
            if 'date' in m else None,
        'width': lambda m: m.get('dimensions', dict()).get('width'),
        'height': lambda m: m.get('dimensions', dict()).get('height'),
        'likescount': lambda m: m.get('likes', dict()).get('count'),
        'commentscount': lambda m: m.get('comments', dict()).get('count'),
        'display_src': lambda m: m.get('display_src'),
        'video_url': lambda m: m.get('video_url'),
    }

    _OWNER_MAP = {}

    def __init__(self, directory=None, profile=None, hashtag=None,
                add_metadata=False, get_videos=False, videos_only=False,
                jobs=16, template="{id}", url_generator=default,
                dump_json=False, dump_only=False, extended_dump=False):
        """Create a new looter instance.

        Keyword Arguments:
            directory (`str`): where downloaded medias will be stored
                **[default: None]**
            profile (`str`): a profile to download media from
                **[default: None]**
            hashtag (`str`): a hashtag to download media from
                **[default: None]**
            add_metadata (`bool`): Add date and comment metadata to
                the downloaded pictures. **[default: False]**
            get_videos (`bool`): Also get the videos from the given
                target **[default: False]**
            videos_only (`bool`): Only download videos (implies
                ``get_videos=True``). **[default: False]**
            jobs (`bool`): the number of parallel threads to use to
                download media (12 or more is advised to have a true parallel
                download of media files) **[default: 16]**
            template (`str`): a filename format, in Python new-style-formatting
                format. See the the Template page of the documentation
                for available keys. **[default: {id}]
            url_generator (`function`): a callable that takes a media
                dictionnary as argument and returs the URL it should
                download the picture from. The default tries to get
                the best available size. **[default: `urlgen.default`]**
            dump_json (`bool`): Save each resource metadata to a
                JSON file next to the actual image/video. **[default: False]**
            dump_only (`bool`): Only save metadata and discard the actual
                resource. **[default: False]**
            extended_dump (`bool`): Attempt to fetch as much metadata as
                possible, at the cost of more time. Set to `True` if, for
                instance, you always want the top comments to be downloaded
                in the dump. **[default: False]**
        """
        if profile is not None and hashtag is not None:
            raise ValueError("Give only a profile or an hashtag, not both !")

        if profile is not None:
            self.target = profile
            self._page_name = 'ProfilePage'
            self._section_name = 'user'
            self._base_url = "https://www.instagram.com/{}/"
        elif hashtag is not None:
            self.target = hashtag
            self._page_name = 'TagPage'
            self._section_name = 'tag'
            self._base_url = "https://www.instagram.com/explore/tags/{}/"
        else:
            self.target = None

        # Create self.directory if it doesn't exist.
        if directory is not None and not os.path.exists(directory):
            os.makedirs(directory)

        self.template = template
        self._required_template_keys = self._RX_TEMPLATE.findall(template)

        self.url_generator = url_generator
        if not callable(url_generator):
            raise ValueError("url_generator must be a callable !")

        self.directory = directory
        self.add_metadata = add_metadata
        self.get_videos = get_videos or videos_only
        self.videos_only = videos_only
        self.dump_json = dump_json or dump_only
        self.dump_only = dump_only
        self.extended_dump = extended_dump
        self.jobs = jobs

        self.session = requests.Session()
        self.session.cookies = six.moves.http_cookiejar.LWPCookieJar(self.COOKIE_FILE)
        load_cookies(self.session)

        self.user_agent = ("Mozilla/5.0 (Windows NT 10.0; WOW64; "  # seems legit
                           "rv:50.0) Gecko/20100101 Firefox/50.0")

        self.dl_count = 0
        self.metadata = {}
        self._workers = []
        self.dl_count_lock = threading.Lock()

        self.session.headers.update({
            'User-Agent': self.user_agent,
            'Accept': 'text/html',
            'Accept-Encoding': 'gzip, deflate',
            'Connection': 'keep-alive',
            'Host': 'www.instagram.com',
            'DNT': '1',
            'Upgrade-Insecure-Requests': '1',
        })

        atexit.register(self.__del__)

    def __del__(self):
        if hasattr(self, 'session'):
            try:
                self.session.close()
            except ReferenceError:
                pass
        if hasattr(self, "_workers"):
            for worker in self._workers:
                worker.kill()

    def __length_hint__(self):
        try:
            data = next(self.pages())['entry_data'][self._page_name][0]
            length = data[self._section_name]['media']['count']
        except (KeyError, StopIteration):
            length = 0
        return length

    def login(self, username, password):
        """Login with provided credentials.

        Arguments:
            username (`str`): the username to log in with
            password (`str`): the password to log in with

        .. note::

            Code taken from LevPasha/instabot.py
        """
        # self.session.cookies.update({
        #     'sessionid': '',
        #     'mid': '',
        #     'ig_pr': '1',
        #     'ig_vw': '1920',
        #     'csrftoken': '',
        #     's_network': '',
        #     'ds_user_id': ''
        # })

        login_post = {'username': username,
                      'password': password}

        self.session.headers.update({
            'Origin': self.URL_HOME,
            'Referer': self.URL_HOME,
            'X-Instragram-AJAX': '1',
            'X-Requested-With': 'XMLHttpRequest',
        })

        with self.session.get(self.URL_HOME) as res:
            self.session.headers.update({'X-CSRFToken': res.cookies['csrftoken']})
        time.sleep(5 * random.random()) # nosec

        with self.session.post(self.URL_LOGIN,
                data=login_post, allow_redirects=True) as login:

            self.session.headers.update({'X-CSRFToken': login.cookies['csrftoken']})
            # save_cookies(self.session, 'cookies.txt')
            if not login.status_code == 200:
                raise SystemError("Login error: check your connection")

        with self.session.get(self.URL_HOME) as res:
            if res.text.find(username) == -1:
                raise ValueError('Login error: check your login data')
            save_cookies(self.session)

    def logout(self):
        """Log out from current session.

        .. note::

            Code taken from LevPasha/instabot.py
        """
        logout_post = {'csrfmiddlewaretoken': self.session.cookies._cookies.get(
            "www.instagram.com", {"/":{}})["/"].get("sessionid") is not None
        }
        self.session.post(self.URL_LOGOUT, data=logout_post)
        if os.path.isfile(self.COOKIE_FILE):
            os.remove(self.COOKIE_FILE)

    def _init_workers(self):
        """Initialize a pool of workers to download files.
        """
        self._shared_map = {}
        self._workers = []
        self._medias_queue = six.moves.queue.Queue()
        for _ in six.moves.range(self.jobs):
            worker = InstaDownloader(self)
            worker.start()
            self._workers.append(worker)

    def pages(self, media_count=None, with_pbar=False):
        """An iterator over the shared data of a profile or hashtag.

        Create a connection to www.instagram.com and use successive
        GET requests to load all pages of a profile. Each page contains
        12 media nodes, as well as metadata associated to the account.

        Arguments:
            media_count (`int`): how many media to show before
                stopping **[default: None]**
            with_pbar (`bool`): display a progress bar. **[default: False]**

        Yields:
            `dict`: the page content deserialised from JSON
        """
        url = self._base_url.format(self.target)
        current_page = 0
        while True:
            current_page += 1
            with self.session.get(url) as res:
                data = self._get_shared_data(res)

            try:
                media_info = data['entry_data'][self._page_name][0][self._section_name]['media']
            except KeyError:
                warnings.warn("Could not find page of user: {}".format(self.target), stacklevel=1)
                return

            if media_count is None:
                media_count = data['entry_data'][self._page_name][0][self._section_name]['media']['count']

            if with_pbar and media_info['page_info']['has_next_page'] and media_info["nodes"]:
                if 'max_id' not in url:  # First page: init pbar
                    self._init_pbar(1, media_count//12 + 1, 'Loading pages |')
                else:  # Other pages: update pbar
                    if self._pbar.value  == self._pbar.max_value:
                        self._pbar.max_value += 1
                    self._pbar.update(self._pbar.value+1)

            # First page: if user page, get metadata
            if 'max_id' not in url and self._section_name=="user":
                self.metadata = self._parse_metadata_from_profile_page(data)

            yield data

            # Break if the page is private (no media to show) or if the last page was reached
            if not media_info['page_info']['has_next_page'] or not media_info['nodes']:

                if not media_info['nodes']:
                    if self._section_name == "tag":
                        msg = "#{} has no medias to show.".format(self.target)
                    elif not self.is_logged_in():
                        msg = "Profile {} is private, retry after logging in.".format(self.target)
                    else:
                        msg = "Profile {} is private, and you are not following it.".format(self.target)
                    warnings.warn(msg)

                break

            else:
                url = '{}?max_id={}'.format(self._base_url.format(self.target), media_info['page_info']["end_cursor"])

    def medias(self, media_count=None, with_pbar=False, timeframe=None):
        """An iterator over the media nodes of a profile or hashtag.

        Using :obj:`InstaLooter.pages`, extract media nodes from each page
        and yields them successively.

        Arguments:
            media_count (`int`): how many media to show before
                stopping **[default: None]**
            with_pbar (`bool`): display a progress bar **[default: False]**
            timeframe (`tuple`): a couple of datetime.date object
                specifying the date frame within which to yield medias
                (a None value can be given as well) **[default: None]**
                **[format: (start, stop), stop older than start]**
        """
        if timeframe is None: # Avoid checking timeframe every loop
            return self._timeless_medias(media_count=media_count, with_pbar=with_pbar)
        else:
            return self._timed_medias(media_count=media_count, with_pbar=with_pbar, timeframe=timeframe)

    def _timeless_medias(self, media_count=None, with_pbar=False):
        seen = set()
        for page in self.pages(media_count=media_count, with_pbar=with_pbar):
            for media in page['entry_data'][self._page_name][0][self._section_name]['media']['nodes']:
                if media['id'] in seen:
                    return
                yield media
                seen.add(media['id'])

    def _timed_medias(self, media_count=None, with_pbar=False, timeframe=None):
        seen = set()
        start_time, end_time = get_times(timeframe)
        for page in self.pages(media_count=media_count, with_pbar=with_pbar):
            for media in page['entry_data'][self._page_name][0][self._section_name]['media']['nodes']:
                media_date = datetime.date.fromtimestamp(media['date'])
                if start_time >= media_date >= end_time:
                    if media['id'] in seen:
                        return
                    yield media
                    seen.add(media['id'])
                elif media_date < end_time:
                    return

    def download_pictures(self, **kwargs):#media_count=None, with_pbar=False, timeframe=None, new_only=False):
        """Download all the pictures from provided target.

        Keyword Arguments:
            media_count (`int`): how many media to download before
                stopping **[default: None]**
            with_pbar (`bool`): display a progress bar **[default: False]**
            timeframe (`tuple`): a couple of datetime.date object
                specifying the date frame within which to yield medias
                (a None value can be given as well) **[default: None]**
                **[format: (start, stop), stop older than start]**
            new_only (`bool`):
                stop looking for new medias if old ones are found in the
                destination directory **[default: False]**
        """
        self.download(_condition=lambda media: not media['is_video'], **kwargs)

    def download_videos(self, **kwargs):#media_count=None, with_pbar=False, timeframe=None):
        """Download all the videos from provided target.

        Keyword Arguments:
            media_count (`int`): how many media to download before
                stopping **[default: None]**
            with_pbar (`bool`): display a progress bar **[default: False]**
            timeframe (`tuple`): a couple of datetime.date object
                specifying the date frame within which to yield medias
                (a None value can be given as well) **[default: None]**
                **[format: (start, stop), stop older than start]**
            new_only (`bool`):
                stop looking for new medias if old ones are found in the
                destination directory **[default: False]**
        """
        self.download(_condition=lambda media: media['is_video'], **kwargs)

    def get_owner_info(self, code):
        """Get metadata about the owner of given post.

        Arguments:
            code (`str`): the code of the post (can be found in the url:
                ``https://www.instagram.com/p/<code>/``) when looking at a
                specific media

        Returns:
            `dict`: the owner metadata deserialised from JSON
        """
        url = "https://www.instagram.com/p/{}/".format(code)
        with self.session.get(url) as res:
            data = self._get_shared_data(res)
        return data['entry_data']['PostPage'][0]['graphql']['shortcode_media']['owner']

    def download(self, **kwargs):
        """Download all the medias from provided target.

        This method follwows the parameters given in the :obj:`__init__`
        method (profile or hashtag, directory, get_videos, videos_only,
        add_metadata and dump_json).

        Keyword Arguments:
            media_count (`int`): how many media to download before
                stopping **[default: None]**
            with_pbar (`bool`): display a progress bar
                **[default: False]**
            timeframe (`tuple`): a couple of :obj:`datetime.date` object
                specifying the date frame within which to yield medias
                (a None value can be given as well) **[default: None]**
                **[format: (start, stop), stop older than start]**
            new_only (`bool`):
                stop looking for new medias if old ones are found in the
                destination directory **[default: False]**
        """
        # extract the parameters here to avoid having a too heavy
        # function signature
        media_count = kwargs.get('media_count', None)
        with_pbar = kwargs.get('with_pbar', False)
        timeframe = kwargs.get('timeframe', None)
        new_only = kwargs.get('new_only', False)

        if self.target is None:
            raise ValueError("No download target was specified during initialisation !")
        elif self.directory is None:
            raise ValueError("No directory was specified during initialisation !")

        self._init_workers()
        if '_condition' not in kwargs:
            if self.videos_only:
                condition = lambda media: media['is_video']
            elif not self.get_videos:
                condition = lambda media: not media['is_video']
            else:
                condition = lambda media: True
        else:
            condition = kwargs.get('_condition')

        medias_queued = self._fill_media_queue(
            media_count=media_count, with_pbar=with_pbar,
            condition=condition, timeframe=timeframe,
            new_only=new_only
        )

        if medias_queued == 0:
            warnings.warn("No {}medias found.".format('new ' if new_only else ''))
        elif with_pbar:
            self._init_pbar(self.dl_count, medias_queued, 'Downloading |')
        self._poison_workers()
        self._join_workers(with_pbar=with_pbar)

    def download_post(self, code):
        """Download a single post referenced by its code.

        Arguments:
            code (`str`): the code of the post (can be found in the url:
                ``https://www.instagram.com/p/<code>/``) when looking
                at a specific media.
        """
        if self.directory is None:
            raise ValueError("No directory was specified during initialisation !")

        self._init_workers()
        media = self.get_post_info(code)

        if media.get('__typename') == "GraphSidecar":
            self._add_sidecars_to_queue(media, lambda m: True, None, 0, False)
        else:
            self._add_media_to_queue(media, lambda m: True, None, 0, False)

        self._poison_workers()
        self._join_workers()

    def get_metadata(self):
        if self._page_name == 'TagPage':
            raise ValueError("Cannot get metadata of an hashtag !")
        elif not self.metadata:
            next(self.pages())
        return self.metadata

    def get_post_info(self, code):
        """Get info about a single post referenced by its code

        Arguments:
            code (`str`): the code of the post (can be found in the url:
                ``https://www.instagram.com/p/<code>/``) when looking
                at a specific media.
        """
        url = "https://www.instagram.com/p/{}/".format(code)
        with self.session.get(url) as res:
        # media = self._get_shared_data(res)['entry_data']['PostPage'][0]['media']
            media = self._get_shared_data(res)['entry_data']['PostPage'][0]\
                                              ['graphql']['shortcode_media']
        # Fix renaming of attributes
        media.setdefault('code', media.get('shortcode'))
        media.setdefault('date', media.get('taken_at_timestamp'))
        media.setdefault('display_src', media.get('display_url'))
        media.setdefault('likes', media['edge_media_preview_like'])
        media.setdefault('comments', media['edge_media_to_comment'])
        return media

    @classmethod
    def _extract_code_from_url(cls, url):
        result = cls._RX_CODE_URL.search(url)
        if result is None:
            raise ValueError("Invalid post url: {}".format(url))
        else:
            return result.group(1)

    def _get_shared_data(self, res):
        soup = bs.BeautifulSoup(res.text, PARSER)
        script = soup.find('body').find('script', {'type': 'text/javascript'})
        return json.loads(self._RX_SHARED_DATA.match(script.text).group(1))

    def _fill_media_queue(self, media_count=None, with_pbar=False, condition=None, timeframe=None, new_only=False):
        medias_queued = 0
        condition = condition or (lambda media: self.get_videos or not media['is_video'])
        for media in self.medias(media_count=media_count, with_pbar=with_pbar, timeframe=timeframe):
            medias_queued, stop = self._add_media_to_queue(media, condition, media_count, medias_queued, new_only)
            if stop:
                break
        return medias_queued

    def _add_media_to_queue(self, media, condition, media_count, medias_queued, new_only):
        if media.get('__typename') == "GraphSidecar":
            return self._add_sidecars_to_queue(
                media, condition, media_count, medias_queued, new_only)
        elif condition(media):
            if self.extended_dump:
                media = self.get_post_info(media.get('shortcode') or media['code'])
            media_basename = self._make_filename(media)
            if not os.path.exists(os.path.join(self.directory, media_basename)):
                medias_queued += 1
                self._medias_queue.put(media)
            # stop here if the file already exists and we want only new files
            elif new_only:
                return medias_queued, True
            # stop here if we have as many files queued as wanted
            if media_count is not None and medias_queued >= media_count:
                return medias_queued, True
        return medias_queued, False

    def _add_sidecars_to_queue(self, media, condition, media_count, medias_queued, new_only):
        media = self.get_post_info(media.get('shortcode') or media['code'])
        for index, sidecar in enumerate(media['edge_sidecar_to_children']['edges']):
            sidecar = self._sidecar_to_media(sidecar['node'], media, index)
            medias_queued, stop = self._add_media_to_queue(
                sidecar, condition, media_count, medias_queued, new_only)
            if stop:
                break
        return medias_queued, stop

    def _make_filename(self, media):

        try:
            media['owner'].update(self._OWNER_MAP[media['owner']['id']])
        except KeyError:
            pass

        required_template_keys = copy.copy(self._required_template_keys)
        if media['is_video']:
            required_template_keys.append('video_url')
        else:
            required_template_keys.append('display_src')

        try:
            template_values = {}
            for x in required_template_keys:
                template_values[x] = value = self._TEMPLATE_MAP[x](media)
                if value is None:
                    raise ValueError
        except ValueError:
            media = self.get_post_info(media.get('code') or media['shortcode'])
            template_values = {x:self._TEMPLATE_MAP[x](media) for x in required_template_keys}
        finally:
            self._OWNER_MAP[media['owner']['id']] = media['owner']

        extension = os.path.splitext(os.path.basename(
            media.get('video_url') or media['display_src']
        ).split('?')[0])[1]

        return "".join([self.template.format(**template_values), extension])

    def _join_workers(self, with_pbar=False):
        while any(w.is_alive() for w in self._workers):
            if with_pbar and hasattr(self, '_pbar'):
                with self.dl_count_lock:
                    self._pbar.update(self.dl_count)
        if with_pbar and hasattr(self, '_pbar'):
            with self.dl_count_lock:
                self._pbar.update(self.dl_count)

    def _init_pbar(self, ini_val, max_val, label):
        self._pbar = progressbar.ProgressBar(
            min_value=0,
            max_value=max_val,
            initial_value=ini_val,
            widgets=[
                label,
                progressbar.Percentage(),
                '(', progressbar.SimpleProgress(), ')',
                progressbar.Bar(),
                progressbar.Timer(), ' ',
                '|', progressbar.ETA(),
            ]
        )
        self._pbar.start()

    def _poison_workers(self):
        for _ in self._workers:
            self._medias_queue.put(None)

    def _parse_metadata_from_profile_page(self, data):
        user = data["entry_data"][self._page_name][0]["user"]
        metadata = {}
        for k, v in six.iteritems(user):
            metadata[k] = copy.copy(v)
        metadata['follows'] = metadata['follows']['count']
        metadata['followed_by'] = metadata['followed_by']['count']
        del metadata['media']['nodes']
        return metadata

    @staticmethod
    def _sidecar_to_media(sidecar, media, index):
        for key in ("owner", "location"):
            sidecar.setdefault(key, media[key])

        # try to get a caption if available
        captions = media['edge_media_to_caption']['edges']
        try:
            sidecar['caption'] = captions[index]['node']['text']
        except IndexError:
            if captions:
                sidecar['caption'] = captions[0]['node']['text']

        sidecar['likes'] = media['edge_media_preview_like']
        sidecar['comments'] = media['edge_media_to_comment']
        sidecar['display_src'] = sidecar['display_url']
        sidecar['code'] = sidecar['shortcode']
        sidecar['date'] = media['taken_at_timestamp']
        return sidecar

    def is_logged_in(self):
        """Check if the current instance is logged in.

        Returns:
            `bool`: if the user is logged in or not
        """
        return self.session.cookies._cookies.get(
            "www.instagram.com", {"/":{}})["/"].get("sessionid") is not None



#!/usr/bin/env python
# coding: utf-8
"""
instaLooter - Another API-less Instagram pictures and videos downloader

Usage:
    instaLooter (login | logout)
    instaLooter batch <batch_file>
    instaLooter <profile> [<directory>] [options]
    instaLooter (hashtag <hashtag> | post <post_token>) <directory> [options]
    instaLooter (-h | --help | --version | --usage)

Arguments:
    <profile>                    The username of the profile to download
                                 videos and pictures from.
    <hashtag>                    A hashtag to download pictures and videos
                                 from.
    <post_token>                 Either the url or the code of a single post
                                 to download the picture or video from.
    <directory>                  The directory in which to download files.
    <batch_file>                 The path to the batch file containing batch
                                 download instructions (see the online
                                 documentation).

Options - Credentials:
    -u USER, --username USER     The username to connect to Instagram with.
    -p PASS, --password PASS     The password to connect to Instagram with
                                 (will be asked in the shell if the `--username`
                                  option was given without the corresponding
                                  `--password`).

Options - Files:
    -n NUM, --num-to-dl NUM      Maximum number of new files to download
    -j JOBS, --jobs JOBS         Number of parallel threads to use to
                                 download files. [default: 16]
    -T TMPL, --template TMPL     A filename template to use to write the
                                 files (see *Template*). [default: {id}]
    -v, --get-videos             Get videos as well as photos.
    -V, --videos-only            Get videos only.
    -N, --new                    Only look for files newer than the ones
                                 in the destination directory (faster).
    -t TIME, --time TIME         The time limit within which to download
                                 pictures and video (see *Time*).

Options - Metadata:
    -m, --add-metadata           Add date and caption metadata to downloaded
                                 pictures (requires PIL/Pillow and piexif).
    -d, --dump-json              Save metadata to a JSON file next to
                                 downloaded videos/pictures.
    -D, --dump-only              Save only the metadata and no video / picture.
    -e, --extended-dump          Always dump the maximum amount of extracted
                                 information, at the cost of more time.

Options - Miscellaneous:
    -q, --quiet                  Do not produce any output.
    -h, --help                   Display this message and quit.
    --version                    Show program version and quit.
    --traceback                  Print error traceback if any (debug).
    -W WARNINGCTL                Change warning behaviour (same as python -W).
                                 [default: default]

Template:
    The default filename of the pictures and videos on Instagram doesn't show
    anything about the file you just downloaded. But using the -t argument
    allows you to give instaLooter a filename template, using the following
    format with brackets-enclosed ({}) variable names among:
    - ``id``*² and ``code``² of the instagram id of the media
    - ``ownerid``*, ``username`` and ``fullname`` of the owner
    - ``datetime``*: the date and time of the post (YYYY-MM-DD hh:mm:ss)
    - ``date``*: the date of the post (YYYY-MM-DD)
    - ``width``* and ``height``*
    - ``likescount``* and ``commentscount``*

    ²: use at least one of these to make sure the generated file name
    is unique (``datetime`` is not unique anymore since multiple-image-posts).

    *: use these only to quicken download, since fetching the others may take
    a tad longer (in particular in hashtag download mode).

    You are however to make sure that the generated filename is unique, so you
    should use at least id, code or datetime somewhere.
    Examples of acceptable values:
        - {username}.{datetime}.{code}
        - {username}-{likescount}-{width}x{height}.{id}

Time:
    The --time parameter can be given either a combination of start and stop
    date in ISO format (e.g. 2016-12-21:2016-12-18, 2015-03-07:, :2016-08-02)
    or a special value among: "thisday", "thisweek", "thismonth", "thisyear".

    Edges are included in the time frame, so if using the following value:
    `--time 2016-05-10:2016-04-03`, then all medias will be downloaded
    including the ones posted the 10th of May 2016 and the 3rd of April 2016.

See more at http://instalooter.readthedocs.io/en/latest/usage.html

"""
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import sys
import getpass
import warnings
import traceback

import six
import hues
import docopt

from . import __version__
from .core import InstaLooter
from .batch import BatchRunner
from .utils import get_times_from_cli, console, wrap_warnings

WARNING_ACTIONS = {
    'error', 'ignore', 'always', 'default', 'module', 'once'
}


def usage():
    return next(s for s in __doc__.split("\n\n") if s.startswith("Usage"))


def login(looter, args):
    if args['--username']:
        username = args['--username']
        if not looter.is_logged_in():
            password = args['--password'] or getpass.getpass()
            looter.login(username, password)
            if not args['--quiet']:
                hues.success('Logged in.')
        elif not args['--quiet']:
            hues.success("Already logged in.")


@wrap_warnings
def main(argv=None):
    """Run from the command line interface.
    """
    argv = argv or sys.argv[1:]

    try:
        args = docopt.docopt(__doc__, argv,
                             version='instaLooter {}'.format(__version__))
    except docopt.DocoptExit as de:
        print(de)
        return 1

    argv_positional = [param for param in argv if not param.startswith("-")]
    if argv_positional[0] in ("post", "hashtag") and len(argv_positional) < 3:
        print(usage())
        return 1

    if args['logout']:
        if not os.path.exists(InstaLooter.COOKIE_FILE):
            hues.error('Cookie file not found.')
            return 1
        InstaLooter().logout()
        hues.success('Logged out.')
        return 0

    elif args['login']:
        try:
            args['--username'] = six.moves.input('Username: ')
            login(InstaLooter(), args)
            return 0
        except ValueError as ve:
            console.error(ve)
            if args["--traceback"]:
               traceback.print_exc()
            return 1

    if args['-W'] not in WARNING_ACTIONS:
        print("Unknown warning action: {}".format(args['-W']))
        print("   available action: {}".format(', '.join(WARNING_ACTIONS)))
        return 1

    if args['batch']:
        with open(args['<batch_file>']) as batch_file:
            batch_runner = BatchRunner(batch_file)
        batch_runner.runAll()
        return 0

    with warnings.catch_warnings():
        warnings.simplefilter(args['-W'])

        # if args['<hashtag>'] and not args['--credentials']:
        #    warnings.warn("#hashtag downloading requires an Instagram account.")
        #    return 1

        if args['<post_token>'] is not None:
            args['--get-videos'] = True

        looter = InstaLooter(
            directory=os.path.expanduser(args.get('<directory>') or os.getcwd()),
            profile=args['<profile>'],hashtag=args['<hashtag>'],
            add_metadata=args['--add-metadata'],
            get_videos=args['--get-videos'],
            videos_only=args['--videos-only'], jobs=int(args['--jobs']),
            template=args['--template'],
            dump_json=args['--dump-json'],
            dump_only=args['--dump-only'],
            extended_dump=args['--extended-dump'],
        )

        try:
            login(looter, args)
            if args['--time']:
                timeframe = get_times_from_cli(args['--time'])
            else:
                timeframe = None
        except ValueError as ve:
            console.error(ve)
            if args["--traceback"]:
               traceback.print_exc()
            return 1

        try:
            post_token = args['<post_token>']
            if post_token is None:
                media_count = int(args['--num-to-dl']) if args['--num-to-dl'] else None
                looter.download(
                    media_count=media_count,
                    with_pbar=not args['--quiet'],
                    timeframe=timeframe,
                    new_only=args['--new'],
                )
            else:
                if 'insta' in post_token:
                    post_token = looter._extract_code_from_url(post_token)
                looter.download_post(post_token)

        except Exception as e:
            console.error(e)
            if args["--traceback"]:
               traceback.print_exc()
            return 1

        except KeyboardInterrupt:
            return 1

        else:
            return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8
from __future__ import (
    absolute_import,
    unicode_literals,
)

import requests
import contextlib
import threading
import datetime
import os
import re
import six
import json

try:
    import PIL.Image
    import piexif
except ImportError:
    PIL = None


class InstaDownloader(threading.Thread):

    @staticmethod
    def _save_metadata(path, metadata):
        """Save metadata in a JSON file named like the resource.
        """
        # replace .jpg / .mp4 extension with .json
        path = re.sub('\..{3}$', '.json', path)
        with open(path, 'w') as fp:
            json.dump(metadata, fp, indent=4, sort_keys=True)

    @staticmethod
    def _get_caption(metadata):
        try:
            return metadata['caption']
        except KeyError:
            pass
        try:
            return metadata['edge_media_to_caption']['edges'][0]['node']['text']
        except KeyError:
            return ''

    def __init__(self, owner):
        super(InstaDownloader, self).__init__()
        self.medias = owner._medias_queue
        self.directory = owner.directory
        self.add_metadata = owner.add_metadata
        self.dump_json = owner.dump_json
        self.dump_only = owner.dump_only
        self.owner = owner
        self.session = requests.Session()
        self.session.cookies = self.owner.session.cookies
        self._killed = False

    def run(self):
        while not self._killed:
            try:
                media = self.medias.get(timeout=1)
                if media is None:
                    break
                elif media.get('is_video'):
                    self._download_video(media)
                else:
                    self._download_photo(media)
                with self.owner.dl_count_lock:
                    self.owner.dl_count += 1
            except six.moves.queue.Empty:
                pass

    def _add_metadata(self, path, metadata):
        """Add some metadata to the picture located at `path`.
        """

        if PIL is not None:

            try:
                full_name = self.owner.metadata['full_name']
            except KeyError:
                full_name = self.owner.get_owner_info(
                    metadata.get('shortcode') or metadata['code']
                )['full_name']

            exif_dict = {"GPS": {}, "thumbnail": None}

            exif_dict['0th'] = {
                piexif.ImageIFD.Artist: \
                    "Image creator, {}".format(full_name).encode('utf-8'),
            }

            exif_dict['1st'] = {
                piexif.ImageIFD.Artist: \
                    "Image creator, {}".format(full_name).encode('utf-8'),
            }

            exif_dict['Exif'] = {
                piexif.ExifIFD.DateTimeOriginal: \
                    datetime.datetime.fromtimestamp(metadata['date']).isoformat(),
                piexif.ExifIFD.UserComment: \
                    self._get_caption(metadata).encode('utf-8'),
            }

            print(exif_dict)

            with PIL.Image.open(path) as img:
                img.save(path, exif=piexif.dump(exif_dict))

    def _download_photo(self, media):
        """Download a picture from a media dictionnary.
        """
        photo_url = self.owner.url_generator(media)

        if not isinstance(photo_url, six.string_types):
            raise RuntimeError('The "custom_photo_url" option must return a string !')

        photo_name = os.path.join(self.directory, self.owner._make_filename(media))

        # save full-resolution photo
        if not self.dump_only:
            self._dl(photo_url, photo_name)

        # put info from Instagram post into image metadata
        if self.add_metadata:
            self._add_metadata(photo_name, media)

        if self.dump_json:
            self._save_metadata(photo_name, media)

    def _download_video(self, media):
        """Download a video from a media dictionnary.
        """
        url = "https://www.instagram.com/p/{}/".format(media.get('shortcode') or media['code'])

        if "video_url" not in media:
            with contextlib.closing(self.session.get(url)) as res:
                # data = self.owner._get_shared_data(res)['entry_data']['PostPage'][0]['media']
                data = self.owner._get_shared_data(res)['entry_data']['PostPage'][0]['graphql']['shortcode_media']
        else:
            data = media

        video_url = data["video_url"]
        #video_basename = os.path.basename(video_url.split('?')[0])
        video_name = os.path.join(self.directory, self.owner._make_filename(data))

        # save video
        if not self.dump_only:
            self._dl(video_url, video_name)

        if self.dump_json:
            self._save_metadata(video_name, media)

    def _dl(self, source, dest):
        """Download a file located at `source` to `dest`.
        """
        self.session.headers['Accept'] = '*/*'
        with contextlib.closing(self.session.get(source)) as res:
            with open(dest, 'wb') as dest_file:
                for block in res.iter_content(1024):
                    if block:
                        dest_file.write(block)

    def kill(self):
        """Kill the Thread.

        This method actually performs a soft kill, it just
        forces the Thread to break the infinite loop. If the Thread
        is currently downloading a file, it will first finish the
        download before exiting.
        """
        self._killed = True



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import six
import getpass

from .core import InstaLooter


class BatchRunner(object):
    """Run `InstaLooter` in batch mode, using a configuration file.
    """

    def __init__(self, handle):
        self.parser = six.moves.configparser.ConfigParser()
        self.parser.readfp(handle) if six.PY2 else self.parser.read_file(handle)

    def _getboolean(self, section_id, key, default=None):
        if self.parser.has_option(section_id, key):
            return self.parser.getboolean(section_id, key)
        return default

    def _getint(self, section_id, key, default=None):
        if self.parser.has_option(section_id, key):
            return self.parser.getint(section_id, key)
        return default

    def _get(self, section_id, key, default=None):
        if self.parser.has_option(section_id, key):
            return self.parser.get(section_id, key)
        return default

    def runAll(self):
        """Run all the jobs specified in the configuration file.
        """
        for section_id in self.parser.sections():
            self.runJob(section_id)

    def runJob(self, section_id):
        """Run a job as described in the section named `section_id`.

        Raises:
            KeyError: when the section could not be found.
        """

        if not self.parser.has_section(section_id):
            raise KeyError('section not found: {}'.format(section_id))

        for target_name in ('users', 'hashtags'):
            targets = self.getTargets(self._get(section_id, target_name))

            for target, directory in six.iteritems(targets):
                looter = InstaLooter(
                    directory=os.path.expanduser(directory),
                    profile=target if target_name=='users' else None,
                    hashtag=target if target_name=='hashtags' else None,
                    add_metadata=\
                        self._getboolean(section_id, 'add-metadata', False),
                    get_videos=\
                        self._getboolean(section_id, 'get-videos', False),
                    videos_only=\
                        self._getboolean(section_id, 'videos-only', False),
                    jobs=self._getint(section_id, 'jobs', 16),
                    template=self._get(section_id, 'template', '{id}'),
                    dump_json=self._getboolean(section_id, 'dump-json', False),
                    dump_only=self._getboolean(section_id, 'dump-only', False),
                    extended_dump=\
                        self._getboolean(section_id, 'extended-dump', False),
                )

                if self.parser.has_option(section_id, 'username'):
                    looter.logout()
                    username = self._get(section_id, 'username')
                    password = self._get(section_id, 'password') or \
                        getpass.getpass('Password for "{}": '.format(username))
                    looter.login(username, password)

                looter.download(
                    media_count=self._getint(section_id, 'num-to-dl'),
                    with_pbar=not self._getboolean(section_id, 'quiet', False),
                    new_only=self._getboolean(section_id, 'new', False),
                )

    def getTargets(self, raw_string):
        """Extract targets from a string in 'key: value' format.
        """
        targets = {}
        if raw_string is not None:
            for line in raw_string.splitlines():
                if line:
                    target, directory = line.split(':')
                    targets[target.strip()] = directory.strip()
        return targets
